/* ------------------------ Beginning Of File: main.cpp ----------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <fstream>
#include "world.hpp"
#include "random.hpp"

#undef SEEK_SET
#undef SEEK_CUR
#undef SEEK_END


#include "mpi.h"

using namespace std;


int main(int argc, char* argv[]) { // main

   ifstream file;
   int size, V, Th, Tc, CPA, iter;
   World gw ;
   
   // We save the jobID (given by the queue management system)
   // It is used to create output files
   // (see script provided for details on how this ID is obtained)
   if(argc>1) strcpy(gw.jobID,argv[1]);

   // usual MPI routines   
   MPI_Init(&argc, &argv);
   int procNum;
   MPI_Comm_rank(MPI_COMM_WORLD,&procNum);
   int tempDEBUG;
   MPI_Comm_size(MPI_COMM_WORLD,&tempDEBUG);

   if(procNum==0) {
      printf("\nMPI launched... on %d nodes\n",tempDEBUG);

      // we open the file containing all the parameters
      file.open("param", ios::in);
   
      // we leave the program if there was an error when opening the file...   
      if(file.bad()) return(1);
   
      // we read the parameters from the file
      file >> size;
      file >> V;
      file >> Th;
      file >> Tc;
      file >> CPA;
      file.close();
      printf("Parameters read...\n");
   }
   
   MPI_Bcast(&size, 1, MPI_INT, 0, MPI_COMM_WORLD);
   MPI_Bcast(&V, 1, MPI_INT, 0, MPI_COMM_WORLD);   
   MPI_Bcast(&Th, 1, MPI_INT, 0, MPI_COMM_WORLD);
   MPI_Bcast(&Tc, 1, MPI_INT, 0, MPI_COMM_WORLD);
   MPI_Bcast(&CPA, 1, MPI_INT, 0, MPI_COMM_WORLD);

   MPI_Barrier(MPI_COMM_WORLD);
   if(procNum==0) {
      printf("Parameters OK...\n");
      printf("size %d, V %d, Th %d, Tc %d, CPA %d\n",size,V,Th,Tc,CPA); 
   }  

   // the time at which the program started is saved
   int time1_=time(0);
   MPI_Barrier(MPI_COMM_WORLD);

   // the world is initialized
   gw.init(size, V, Th, Tc, CPA);
   MPI_Barrier(MPI_COMM_WORLD);
   if(procNum==0) printf("World initialized...\n");
   
   // waiting for all processes to reach this point
   MPI_Barrier(MPI_COMM_WORLD);

   // the time at which the simulation started is saved
   int time2_=time(0);

   // the simulation starts
   iter=gw.run();

   // runtime statistics
   int time3_=time2_-time1_;
   int time4_=time(0)-time2_;
   double ave=(double)iter/time4_;
   if(procNum==0) {
      printf("\nExecution time: %d\n",time4_+time3_);
      printf("Iterations: %d\nAverage: %f it/s\n",iter,ave);
   }

   MPI_Finalize();
   return 0 ;
}

/* ------------------------------ End Of File: main.cpp ------------------------------- */
