/* ------------------------ Beginning Of File: node.cpp ----------------------- */

#include "node.hpp"
#include "agent.hpp"
#include "random.hpp"
#include "world.hpp"
// other includes are in node.hpp

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cmath>
#include <algorithm>

// definition of the various agents' life expectancy
// VALUES ARE TO BE CHECKED AND REFERENCED

// virion, life expectancy = 3 days = 5514.89 iterations
#define MAXAGEV 5515
// Th, life expectancy = 1 year = 670978.72 iterations
#define MAXAGETH (int)random_gauss(670979,10000)
// Tc, life expectancy = 1 years = 670978.72 iterations
#define MAXAGETC (int)random_gauss(670979,10000)
// APC, life expectancy = 1 year = 670978.72 iterations
#define MAXAGECPA (int)random_gauss(670979,10000)

// when the simulation is starting, the agents are already "old"
#define MAXAGEV_DEB random_int(1,MAXAGEV)
#define MAXAGETH_DEB random_int(1,MAXAGETH)
#define MAXAGETC_DEB random_int(1,MAXAGETC)
#define MAXAGECPA_DEB random_int(1,MAXAGECPA)

// definition of the maximum number of agents in each matrix element
#define NBVIRUSCELL 20
#define NBTHCELL 10
#define NBTCCELL 10
#define NBCPACELL 10

#define NBNOEUDS numberOfNodes

#define STRAIN_NUMBER 500
#define EPITOPE_NUMBER 1000

// a few global variables used for validation
int _CPA_act = 0;
int _Tc_act = 0;
int _Tc_cree = 0;
int _V_killed =0;
int _V_created =0;


/***************************************************************
*                                                              *
*    Node::Node()                                              *
*          constructor                                         *
*                                                              *
***************************************************************/

Node::Node() {
   myMatrix_  = NULL ;
}



/***************************************************************
*                                                              *
*    Node::Node()                                              *
*          destructor                                          *
*                                                              *
***************************************************************/

Node::~Node() {
   // Desallocation de la matrice
   if( myMatrix_ )
      delete [] myMatrix_ ;
}


/***************************************************************
*                                                              *
*    Node::initStrains()                                       *
*	   initializes the strain array in the current node        *
*                                                              *
***************************************************************/

void Node::initStrains() {
   for(int i=0;i<STRAIN_NUMBER;++i)
      strains[i].init(i);
}



/***************************************************************
*                                                              *
*    Node::init()                                              *
*	   initializes the agent population                        *
*                                                              *
***************************************************************/

bool Node::init(int inSize,int inNbVirus,int inNbTh,int inNbTc,int inNbCPA) {
   int i, initI, initJ;
   Cellule *ptrCell;

   // Initialization of the matrix size for the nodes
   size_ = inSize;

   // Initialization of "global" variables
   nbV_= inNbVirus;
   nbTh_= inNbTh; nbThInf_=0; nbThAct_=0;
   nbTc_= inNbTc; nbTcAct_=0;
   nbCPA_= inNbCPA; nbCPAPres_=0;

   nbV_Init= inNbVirus;
   nbTh_Init= inNbTh;
   nbTc_Init= inNbTc;
   nbCPA_Init= inNbCPA;

   nbVCrea=0; nbVKill=0; nbVRec=0; nbVSent=0;
   nbThCrea=0; nbThKill=0; nbThRec=0; nbThSent=0;
   nbTcCrea=0; nbTcKill=0; nbTcRec=0; nbTcSent=0;
   nbCPACrea=0; nbCPAKill=0; nbCPARec=0; nbCPASent=0;

   nbThInf=0; nbThAct=0; nbTcAct=0; nbCPAPres=0;
   nbTcVerif_=0;nbCPAVerif_=0;

   // Matrix allocation
   myMatrix_ = new Cellule [ size_*size_ ];

   V_tab = new Virus [size_*size_*NBVIRUSCELL];
   for(i=0;i<size_*size_*NBVIRUSCELL;++i) V_free.push_back(i);
   
   Th_tab = new Th [size_*size_*NBTHCELL];
   for(i=0;i<size_*size_*NBTHCELL;++i) Th_free.push_back(i);
   
   Tc_tab = new Tc [size_*size_*NBTCCELL];
   for(i=0;i<size_*size_*NBTCCELL;++i) Tc_free.push_back(i);
   
   CPA_tab = new CPA [size_*size_*NBCPACELL];
   for(i=0;i<size_*size_*NBCPACELL;++i) CPA_free.push_back(i);

   // If the matrix was not allocated properly, the program stops
   if( !myMatrix_ || !V_tab || !Th_tab || !Tc_tab || !CPA_tab) {
      fprintf( stderr, "Allocation error.\n" );
      return false ;
   }
   
   // Initialization of "local" variables
   for(i=0;i<size_*size_;++i) {
      myMatrix_[i].setNbcV(0);
      myMatrix_[i].setNbcTh(0);
      myMatrix_[i].setNbcTc(0);
      myMatrix_[i].setNbcCPA(0);    
   }

   // Virions are randomly distributed on the matrix
   i=0;
   while(i<inNbVirus) {
      // A matrix element is randomly chosen
      initI = random_int(0,size_-1); // Random integer between 0 and size_-1
      initJ = random_int(0,size_-1);
      ptrCell = access( initI, initJ );
      // Parameters are set and the agent is put in the matrix element
      if(ptrCell->getNbcV()<NBVIRUSCELL) {
         int V_loc=V_free.back();
         V_free.pop_back();
         ptrCell->tabV[ptrCell->getNbcV()]=V_loc;
         V_tab[V_loc].setStrain(random_int(0,STRAIN_NUMBER-1));
         V_tab[V_loc].setDate(0);
         V_tab[V_loc].setDest(-1);
         V_tab[V_loc].setMaxAge(MAXAGEV_DEB);
         ptrCell->nbcVInc();
         ++i;_V_created++;
      }
   }

   // Th agents are randomly distributed on the matrix
   i=0;
   while(i<inNbTh) {
      // A matrix element is randomly chosen
      initI = random_int(0,size_-1);
      initJ = random_int(0,size_-1);
      ptrCell = access( initI, initJ );
      // Parameters are set and the agent is put in the matrix element
      if(ptrCell->getNbcTh()<NBTHCELL) {
         int Th_loc=Th_free.back();
         Th_free.pop_back();
         ptrCell->tabTh[ptrCell->getNbcTh()]=Th_loc;
         Th_tab[Th_loc].setClonotype(random_int(0,EPITOPE_NUMBER));
         
         // Standard init, not accounting for GI tract
         if(random_int(0,100)<=40) Th_tab[Th_loc].setActive(-1);
         else Th_tab[Th_loc].setActive(0);
         if(random_int(0,100)<=5) Th_tab[Th_loc].setMemory(-1);
         else Th_tab[Th_loc].setMemory(0);


         // Refined init, which allows inclusion of the GI tract
         // To use this one:
         //    - remove commenting,
         //    - specify which nodes to use for GI tract (in IF statement)
         //    - comment the standard init

/*
         // nodes dedicated to GI tract implementation
         if((procNum>=11 && procNum<=14)||(procNum>=17 && procNum<=22)) {
            if(random_int(0,100)<=85) Th_tab[Th_loc].setActive(-1);
            else Th_tab[Th_loc].setActive(0);
            if(random_int(0,100)<=60) Th_tab[Th_loc].setMemory(-1);
            else Th_tab[Th_loc].setMemory(0);
         }
         
         // "standard" nodes
         else {
            if(random_int(0,100)<=40) Th_tab[Th_loc].setActive(-1);
            else Th_tab[Th_loc].setActive(0);
            if(random_int(0,100)<=5) Th_tab[Th_loc].setMemory(-1);
            else Th_tab[Th_loc].setMemory(0);
         }
*/

         Th_tab[Th_loc].setInfected(0);
         Th_tab[Th_loc].setExpansion(0);
         Th_tab[Th_loc].setDate(0);
         Th_tab[Th_loc].setDest(-1);
         Th_tab[Th_loc].setMaxAge(MAXAGETH_DEB);
         ptrCell->nbcThInc();
         ++i;
      }
   }
   
   // Tc agents are randomly distributed on the matrix
   i=0;
   while(i<inNbTc) {
      // A matrix element is randomly chosen
      initI = random_int(0,size_-1);
      initJ = random_int(0,size_-1);
      ptrCell = access( initI, initJ );
      // Parameters are set and the agent is put in the matrix element
      if(ptrCell->getNbcTc()<NBTCCELL) {
         int Tc_loc=Tc_free.back();
         Tc_free.pop_back();
         ptrCell->tabTc[ptrCell->getNbcTc()]=Tc_loc;
         Tc_tab[Tc_loc].setActive(0);
         Tc_tab[Tc_loc].setExpansion(0);
         Tc_tab[Tc_loc].setClonotype(random_int(0,EPITOPE_NUMBER));
         Tc_tab[Tc_loc].setDate(0);
         Tc_tab[Tc_loc].setDest(-1);

         // Standard init, not accounting for GI tract
         if(random_int(0,1000)<=5) Tc_tab[Tc_loc].setMemory(random_int(0,STRAIN_NUMBER));
         else Tc_tab[Tc_loc].setMemory(0);


         // Refined init, which allows inclusion of the GI tract
         // To use this one:
         //    - remove commenting,
         //    - specify which nodes to use for GI tract (in IF statement)
         //    - comment the standard init

/*
         // nodes dedicated to GI tract implementation
         if((procNum>=11 && procNum<=14)||(procNum>=17 && procNum<=22)) {
            if(random_int(0,100)<=50) {
               Tc_tab[Tc_loc].setMemory(random_int(0,STRAIN_NUMBER));
            }
            else Tc_tab[Tc_loc].setMemory(0);
         }

         // "standard" nodes
         else {
            if(random_int(0,1000)<=5) {
               Tc_tab[Tc_loc].setMemory(random_int(0,STRAIN_NUMBER));
            }
            else Tc_tab[Tc_loc].setMemory(0);
         }
*/

         Tc_tab[Tc_loc].setMaxAge(MAXAGETC_DEB);
         ptrCell->nbcTcInc();
         ++i;

      }
   }
   
   // APC agents are randomly distributed on the matrix
   i=0;
   while(i<inNbCPA) {
      // A matrix element is randomly chosen
      initI = random_int(0,size_-1);
      initJ = random_int(0,size_-1);
      ptrCell = access( initI, initJ );
      // Parameters are set and the agent is put in the matrix element
      if(ptrCell->getNbcCPA()<NBCPACELL) {
         int CPA_loc=CPA_free.back();
         CPA_free.pop_back();
         ptrCell->tabCPA[ptrCell->getNbcCPA()]=CPA_loc;
         CPA_tab[CPA_loc].setPresenting1(0);
         CPA_tab[CPA_loc].setPresenting2(0);
         CPA_tab[CPA_loc].setPresenting3(0);
         CPA_tab[CPA_loc].setPresenting4(0);
         CPA_tab[CPA_loc].setPresenting5(0);
         CPA_tab[CPA_loc].setDate(0);
         CPA_tab[CPA_loc].setDest(-1);
         CPA_tab[CPA_loc].setMaxAge(MAXAGECPA_DEB);
         ptrCell->nbcCPAInc();      
         ++i;
      }
   }
   
   return true ;
}



/***************************************************************
*                                                              *
*    Node::moveV()                                             *
*	   moves a virion from a matrix element to another         *
*                                                              *
***************************************************************/

void Node::moveV(int V_loc, Cellule *dep, Cellule *arr) {

   arr->tabV[arr->getNbcV()]=V_loc;

   // The matrix element it just left is updated
   dep->flushV(V_loc);
   
   // The local counts of agents are updated
   dep->nbcVDec();
   arr->nbcVInc();

}



/***************************************************************
*                                                              *
*    Node::moveTh()                                            *
*	   moves a Th agent from a matrix element to another       *
*                                                              *
***************************************************************/

void Node::moveTh(int Th_loc, Cellule *dep, Cellule *arr) {

   arr->tabTh[arr->getNbcTh()]=Th_loc;
   
   // The matrix element it just left is updated
   dep->flushTh(Th_loc);
   
   // The local counts of agents are updated
   dep->nbcThDec();
   arr->nbcThInc();

}



/***************************************************************
*                                                              *
*    Node::moveTc()                                            *
*	   moves a Tc agent from a matrix element to another       *
*                                                              *
***************************************************************/

void Node::moveTc(int Tc_loc, Cellule *dep, Cellule *arr) {

   arr->tabTc[arr->getNbcTc()]=Tc_loc;

   // The matrix element it just left is updated
   dep->flushTc(Tc_loc);
   
   // The local counts of agents are updated
   dep->nbcTcDec();
   arr->nbcTcInc();

}



/***************************************************************
*                                                              *
*    Node::moveCPA()                                           *
*	   moves an APC agent from a matrix element to another     *
*                                                              *
***************************************************************/

void Node::moveCPA(int CPA_loc, Cellule *dep, Cellule *arr) {

   arr->tabCPA[arr->getNbcCPA()]=CPA_loc;

   // The matrix element it just left is updated
   dep->flushCPA(CPA_loc);
   
   // The local counts of agents are updated
   dep->nbcCPADec();
   arr->nbcCPAInc();

}



/***************************************************************
*                                                              *
*    Node::iteration()                                         *
*	   main function for the node                              *
*          computes a time step inside the node                *
*                                                              *
***************************************************************/

void Node::iteration(int current, World* wd, int node) {
   int i, j, cpt, newCell,temp, tempRandom;
   int agentDepartSuccess, presenting;
   int numberOfNodes=wd->numberOfNodes;
   int ptrV, ptrTh, ptrThVois, ptrTc, ptrTcVois, ptrCPA;
   Cellule *ptrCell, *ptrCellVois;
   move_t move;

   // The update loop is run for each matrix element
   for(cpt=0;cpt<size_*size_;++cpt) { // IF_388

      // The pointer is set on the current matrix element
      ptrCell=&myMatrix_[cpt];

      // Simulating an action from one virion of the matrix element
      if(ptrCell->getNbcV()>0) { // IF_394

         // The pointer is set on a virion randomly chosen
         ptrV=ptrCell->getV();

         // If the agent is still alive, it can move
         if(current-V_tab[ptrV].getDate()<V_tab[ptrV].getMaxAge()) {  // IF_400

            // The next possible destination is computed
            i = getI(cpt); j = getJ(cpt);
            move=V_tab[ptrV].getMove();
            i = getMoveI( i, move ); j = getMoveJ( j, move );
            newCell=(i*size_+j)%(size_*size_);
            if(newCell<0) newCell+=(size_*size_);
            ptrCellVois=&myMatrix_[newCell];

            // The agent is moved if there is enough room left
            if(ptrCellVois->getNbcV()<NBVIRUSCELL)
               moveV(ptrV,ptrCell,ptrCellVois);

            // The agent leaves the node if it moved to the exit point
            if(newCell==(size_*size_-1)) { // IF_414

               // The agent is added to the list of agents leaving the node
               // The second parameter is the type of agent (0 is for virions)
               if(V_tab[ptrV].getDest()<0) V_tab[ptrV].setDest(random_int(0,NBNOEUDS-1));
               tempRandom=random_int(0,100);

               if((wd->sendTo[1]>=0)&&((wd->appartient(wd->route,V_tab[ptrV].getDest()))||(tempRandom>=50)))
                  agentDepartSuccess=wd->addMovingAgent_list2(V_tab[ptrV].getDest(),0,
                               V_tab[ptrV].getDate(),
                               V_tab[ptrV].getMaxAge(),
                               V_tab[ptrV].getStrain());
               else
                  agentDepartSuccess=wd->addMovingAgent_list1(V_tab[ptrV].getDest(),0,
                               V_tab[ptrV].getDate(),
                               V_tab[ptrV].getMaxAge(),
                               V_tab[ptrV].getStrain());

               // The exit point is updated
               if(agentDepartSuccess==0) {
                  ptrCellVois->flushV(ptrV);
                  ptrCellVois->nbcVDec();
                  nbV_--;nbVSent++;
                  V_free.push_back(ptrV);
               }
            } // ENDIF_414
         } // ENDIF_400
	 
         // If the agent is too old, it dies
         else {
            ptrCell->flushV(ptrV);
            ptrCell->nbcVDec();
            nbV_--;nbVKill++;
            V_free.push_back(ptrV);
         }
      } // ENDIF_394

      // Simulating an action from one Th agent of the matrix element
      if(ptrCell->getNbcTh()>0) { // IF_447
         // The pointer is set on a Th agent randomly chosen
         ptrTh=ptrCell->getTh();

         // If the agent is still alive, it can move and perform actions
         if(current-Th_tab[ptrTh].getDate()<Th_tab[ptrTh].getMaxAge()) { // IF_453
            // The next possible destination is computed
            i = getI(cpt); j = getJ(cpt);
            move=Th_tab[ptrTh].getMove();
            i = getMoveI( i, move ); j = getMoveJ( j, move );
            newCell=(i*size_+j)%(size_*size_);
            if(newCell<0) newCell+=(size_*size_);
            ptrCellVois=&myMatrix_[newCell];
            // The agent is moved if there is enough room left
            if(ptrCellVois->getNbcTh()<NBTHCELL) {
               moveTh(ptrTh,ptrCell,ptrCellVois);
               ptrCell=&myMatrix_[newCell%(size_*size_)];
            }
            else newCell=cpt;
            // The agent leaves the node if it moved to the exit point
            // The second parameter is the type of agent (1 is for Th agents)
            if(newCell==(size_*size_-1)) { // IF_471
               if(Th_tab[ptrTh].getDest()<0) {
                  if(random_int(0,100)>=50)
                     // lymphocytes preferentially recirculate to the tissue of origin
                     // (see Thesis)
                     Th_tab[ptrTh].setDest(procNum);
                  else
                     Th_tab[ptrTh].setDest(random_int(0,NBNOEUDS-1));
               }
               tempRandom=random_int(0,100);
               if((wd->sendTo[1]>=0)&&((wd->appartient(wd->route,Th_tab[ptrTh].getDest()))||(tempRandom>=50)))
                  agentDepartSuccess=wd->addMovingAgent_list2(Th_tab[ptrTh].getDest(),1,
                                 Th_tab[ptrTh].getDate(),
                                 Th_tab[ptrTh].getMaxAge(),
                                 Th_tab[ptrTh].getClonotype(),
                                 Th_tab[ptrTh].getActive(),				   
                                 Th_tab[ptrTh].getInfected(),
                                 Th_tab[ptrTh].getMemory(),
                                 Th_tab[ptrTh].getExpansion());
               else
                  agentDepartSuccess=wd->addMovingAgent_list1(Th_tab[ptrTh].getDest(),1,
                                 Th_tab[ptrTh].getDate(),
                                 Th_tab[ptrTh].getMaxAge(),
                                 Th_tab[ptrTh].getClonotype(),
                                 Th_tab[ptrTh].getActive(),				   
                                 Th_tab[ptrTh].getInfected(),
                                 Th_tab[ptrTh].getMemory(),
                                 Th_tab[ptrTh].getExpansion());

               // The exit point is updated
               if(agentDepartSuccess==0) {
                  ptrCellVois->flushTh(ptrTh);
                  ptrCellVois->nbcThDec();
                  nbTh_--;nbThSent++;
                  Th_free.push_back(ptrTh);
               }
            } // ENDIF_471
	    
            // If it moved to a standard element, it can perform actions
            else { // IF_487

               // A Th agent in its expansion phase produces other Tc agents
               if((current-Th_tab[ptrTh].getDate()>0)&&(random_int(0,4700)<=100)) { // IF_616
                  if(Th_tab[ptrTh].getExpansion()>0) {

                     // The new agent is produced in a neighbouring element
                     int decal=0, thereIsNoRoom=0;
                     do {
                        if(newCell-decal<0) decal-=(size_*size_);
                        ptrCellVois=&myMatrix_[newCell-decal];
                        ++decal; ++thereIsNoRoom;
                     }
                     while(ptrCellVois->getNbcTh()==NBTHCELL && thereIsNoRoom<size_*size_);

	                 if(thereIsNoRoom<size_*size_) {
                        ptrThVois=Th_free.back();
                        Th_free.pop_back();
                        ptrCellVois->tabTh[ptrCellVois->getNbcTh()]=ptrThVois;
                        Th_tab[ptrThVois].setActive(Th_tab[ptrTh].getActive());
                        Th_tab[ptrThVois].setExpansion(0);
                        Th_tab[ptrThVois].setInfected(0);
                        Th_tab[ptrThVois].setClonotype(Th_tab[ptrTh].getClonotype());
                        Th_tab[ptrThVois].setDate(current);
                        Th_tab[ptrThVois].setDest(-1);
                        Th_tab[ptrThVois].setMemory(0);
                        Th_tab[ptrThVois].setMaxAge(MAXAGETC);
                        ptrCellVois->nbcThInc();
                        Th_tab[ptrTh].setExpansion((Th_tab[ptrTh].getExpansion()+1)%5);
                        nbTh_++;nbThCrea++;
                        nbThAct_++;

                        // Re-activated memory cells produce twice as many cells, and faster
                        if(Th_tab[ptrTh].getMemory()==1) {
                           int decal=0;
                           thereIsNoRoom=0;
                           do {
                              if(newCell-decal<0) decal-=(size_*size_);
                              ptrCellVois=&myMatrix_[newCell-decal];
                              ++decal; ++thereIsNoRoom;
                           }
                           while(ptrCellVois->getNbcTh()==NBTHCELL && thereIsNoRoom<size_*size_);
                           if(thereIsNoRoom<size_*size_) {
                              ptrThVois=Th_free.back();
                              Th_free.pop_back();
                              ptrCellVois->tabTh[ptrCellVois->getNbcTh()]=ptrThVois;
                              Th_tab[ptrThVois].setActive(Th_tab[ptrTh].getActive());
                              Th_tab[ptrThVois].setExpansion(0);
                              Th_tab[ptrThVois].setInfected(0);
                              Th_tab[ptrThVois].setClonotype(Tc_tab[ptrTc].getClonotype());
                              Th_tab[ptrThVois].setDate(current);
                              Th_tab[ptrThVois].setDest(-1);
                              Th_tab[ptrThVois].setMemory(0);
                              Th_tab[ptrThVois].setMaxAge(MAXAGETC);
                              ptrCellVois->nbcThInc();
                              Th_tab[ptrTh].setExpansion((Tc_tab[ptrTc].getExpansion()+1)%5);
                              nbTh_++; nbThCrea++;
                              nbThAct_++;
                           }
                        }
                     }
                  }
               } // ENDIF_616

               // An infected Th may produce a new virion
               if(Th_tab[ptrTh].getInfected()>0) { //IF_490
                  int prodRate=strains[V_tab[ptrV].getStrain()].getProdRate();
                  if((current-Th_tab[ptrTh].getDate()>0)&&(random_int(0,50000)<=prodRate)) {

                     // The new virion is produced in a neighbouring element
                    int decal=0, thereIsNoRoom=0;
                     do {
                        if((newCell-decal)<0) decal-=(size_*size_);
                        ptrCellVois=&myMatrix_[newCell-decal];
                        ++decal; ++thereIsNoRoom;
                     }
                     while(ptrCellVois->getNbcV()==NBVIRUSCELL && thereIsNoRoom<(size_*size_/10));
                     if(thereIsNoRoom<(size_*size_/10)) {

                        ptrV=V_free.back();
                        V_free.pop_back();
                        ptrCellVois->tabV[ptrCellVois->getNbcV()]=ptrV;
                        V_tab[ptrV].setStrain(Th_tab[ptrTh].getInfected());
                        V_tab[ptrV].setDate(current);
                        V_tab[ptrV].setDest(-1);
                        V_tab[ptrV].setMaxAge(MAXAGEV);
                        ptrCellVois->nbcVInc();
                        nbV_++;nbVCrea++;
                     }
                  }
               } // ENDIF_490

               // A non-infected Th can get infected
               else { // IF_518
                  if((Th_tab[ptrTh].getActive()!=0)&&(ptrCell->getNbcV()>0)) {
                     ptrV=ptrCell->getV();
                     // A mutation may occur
                     int newStrain=strains[V_tab[ptrV].getStrain()].mutation();
                     // Th is infected if the mutation was successful
                     if(newStrain>=0) {
                        Th_tab[ptrTh].setInfected(newStrain);
                        Th_tab[ptrTh].setMaxAge(current+2500); // infected Th cells have a reduced life span
                        nbThInf++;
                     }
                     // The virion is therefore destroyed
                     nbV_--;nbVKill++;
                     ptrCell->flushV(ptrV);
                     ptrCell->nbcVDec();
                  }
               } // ENDIF_518

               // A passive Th can get activated by APC presenting an antigen
               if((Th_tab[ptrTh].getActive()==0)&&(ptrCell->getNbcCPA()>0)) { // IF_532
                  ptrCPA=ptrCell->getCPA();
                  presenting=CPA_tab[ptrCPA].getPresenting();
                  if(presenting>0) {
                     // The activation is only taking place if the Th recognizes the strain
                     if(VirusRecog(presenting,Th_tab[ptrTh].getClonotype())) {
                        Th_tab[ptrTh].setActive(presenting);
                        nbThAct++;
                        // Infection can occur during presentation
                        if(Th_tab[ptrTh].getInfected()==0 && random_int(0,100)<=40) {
                           Th_tab[ptrTh].setInfected(presenting);
                           nbThInf++;
                        }
                     }
                  }
               } // ENDIF_532

               // A Th can also be activated by another foreign invader
               if(Th_tab[ptrTh].getActive()==0) {
                  int rand=random_int(0,1000);
                  if(rand<8) Th_tab[ptrTh].setActive(-1);
               }
            } // ENDIF_487
         } // ENDIF_453
	 
         // If the agent is too old, it dies
         else {
            if(Th_tab[ptrTh].getActive()>0) nbThAct_--;
            if(Th_tab[ptrTh].getInfected()>0) nbThInf_--;
            nbTh_--;nbThKill++;
            ptrCell->flushTh(ptrTh);
            ptrCell->nbcThDec();
            Th_free.push_back(ptrTh);
         }
      } //ENDIF_447

      // Simulating an action from one Tc agent of the matrix element
      if(ptrCell->getNbcTc()>0) { // IF_570
         // The pointer is set on a Tc agent randomly chosen
         ptrTc=ptrCell->getTc();

         // If the agent is still alive, it can move and perform actions
         if(current-Tc_tab[ptrTc].getDate()<Tc_tab[ptrTc].getMaxAge()) { //IF_576

            // The next possible destination is computed
            i = getI(cpt); j = getJ(cpt);
            move=Tc_tab[ptrTc].getMove();
            i = getMoveI( i, move ); j = getMoveJ( j, move );
            newCell=(i*size_+j)%(size_*size_);
            if (newCell<0) newCell+=(size_*size_);
            ptrCellVois=&myMatrix_[newCell];
	    
            // The agent is moved if there is enough room left
            if(ptrCellVois->getNbcTc()<NBTCCELL) {
               moveTc(ptrTc,ptrCell,ptrCellVois);
               ptrCell=&myMatrix_[newCell%(size_*size_)];
            }
            else newCell=cpt;

            // The agent leaves the node if it moved to the exit point,
            // (but only with a 50% probability: this is to respect the fact that
            // the CD4/CD8 ratio in recirculating cells is about twice as important)
            // The second parameter is the type of agent (2 is for Tc agents)
            if(newCell==(size_*size_-1) && random_int(0,100)>=50) { // IF_596
               if(Tc_tab[ptrTc].getDest()<0) {
                  if(random_int(0,100)>=50)
                     // lymphocytes preferentially recirculate to the tissue of origin
                     // (see Thesis)
                     Tc_tab[ptrTc].setDest(procNum);
                  else
                     Tc_tab[ptrTc].setDest(random_int(0,NBNOEUDS-1));
               }
               tempRandom=random_int(0,100);
               if((wd->sendTo[1]>=0)&&((wd->appartient(wd->route,Tc_tab[ptrTc].getDest()))||(tempRandom>=50)))
                  agentDepartSuccess=wd->addMovingAgent_list2(Tc_tab[ptrTc].getDest(),2,
                              Tc_tab[ptrTc].getDate(),
                              Tc_tab[ptrTc].getMaxAge(),
                              Tc_tab[ptrTc].getClonotype(),
                              Tc_tab[ptrTc].getActive(),				   
                              Tc_tab[ptrTc].getExpansion(),
                              Tc_tab[ptrTc].getMemory());
               else
                  agentDepartSuccess=wd->addMovingAgent_list1(Tc_tab[ptrTc].getDest(),2,
                              Tc_tab[ptrTc].getDate(),
                              Tc_tab[ptrTc].getMaxAge(),
                              Tc_tab[ptrTc].getClonotype(),
                              Tc_tab[ptrTc].getActive(),				   
                              Tc_tab[ptrTc].getExpansion(),
                              Tc_tab[ptrTc].getMemory());

               // The exit point is updated
               if(agentDepartSuccess==0) {
                  ptrCellVois->flushTc(ptrTc);
                  ptrCellVois->nbcTcDec();
                  nbTc_--;nbTcSent++;
                  Tc_free.push_back(ptrTc);
               }
            } //ENDIF_596
	    
            // If it moved to a standard element, it can perform actions
            else { //IF_613

               // A Tc agent in its expansion phase produces other Tc agents
               if((current-Tc_tab[ptrTc].getDate()>0)&&(random_int(0,4700)<=100)) { // IF_616
                  if(Tc_tab[ptrTc].getExpansion()>0) {

                     // The new agent is produced in a neighbouring element
                     int decal=0, thereIsNoRoom=0;
                     do {
                        if(newCell-decal<0) decal-=(size_*size_);
                        ptrCellVois=&myMatrix_[newCell-decal];
                        ++decal; ++thereIsNoRoom;
                     }
                     while(ptrCellVois->getNbcTc()==NBTCCELL && thereIsNoRoom<size_*size_);

	                 if(thereIsNoRoom<size_*size_) {
                        ptrTcVois=Tc_free.back();
                        Tc_free.pop_back();
                        ptrCellVois->tabTc[ptrCellVois->getNbcTc()]=ptrTcVois;
                        Tc_tab[ptrTcVois].setActive(Tc_tab[ptrTc].getActive());
                        Tc_tab[ptrTcVois].setExpansion(0);
                        Tc_tab[ptrTcVois].setClonotype(Tc_tab[ptrTc].getClonotype());
                        Tc_tab[ptrTcVois].setDate(current);
                        Tc_tab[ptrTcVois].setDest(-1);
                        Tc_tab[ptrTcVois].setMemory(0);
                        Tc_tab[ptrTcVois].setMaxAge(MAXAGETC);
                        ptrCellVois->nbcTcInc();
                        Tc_tab[ptrTc].setExpansion((Tc_tab[ptrTc].getExpansion()+1)%5);
                        nbTc_++;nbTcCrea++;
                        nbTcAct_++;

                        // Re-activated memory cells produce twice as many cells, and faster
                        if(Tc_tab[ptrTc].getMemory()==1) {
                           int decal=0;
                           thereIsNoRoom=0;
                           do {
                              if(newCell-decal<0) decal-=(size_*size_);
                              ptrCellVois=&myMatrix_[newCell-decal];
                              ++decal; ++thereIsNoRoom;
                           }
                           while(ptrCellVois->getNbcTc()==NBTCCELL && thereIsNoRoom<size_*size_);
                           if(thereIsNoRoom<size_*size_) {
                              ptrTcVois=Tc_free.back();
                              Tc_free.pop_back();
                              ptrCellVois->tabTc[ptrCellVois->getNbcTc()]=ptrTcVois;
                              Tc_tab[ptrTcVois].setActive(Tc_tab[ptrTc].getActive());
                              Tc_tab[ptrTcVois].setExpansion(0);
                              Tc_tab[ptrTcVois].setClonotype(Tc_tab[ptrTc].getClonotype());
                              Tc_tab[ptrTcVois].setDate(current);
                              Tc_tab[ptrTcVois].setDest(-1);
                              Tc_tab[ptrTcVois].setMemory(0);
                              Tc_tab[ptrTcVois].setMaxAge(MAXAGETC);
                              ptrCellVois->nbcTcInc();
                              Tc_tab[ptrTc].setExpansion((Tc_tab[ptrTc].getExpansion()+1)%5);
                              nbTc_++; nbTcCrea++;
                              nbTcAct_++;
                           }
                        }
                     }
                  }
               } // ENDIF_616

               // An active Tc can destroy virions and infected cells
               if(Tc_tab[ptrTc].getActive()>0) { // IF_653
                  if(ptrCell->getNbcCPA()>0) {
                     ptrCPA=ptrCell->getCPA();
                     if(CPA_tab[ptrCPA].getPresenting()==Tc_tab[ptrTc].getActive()) {
                        ptrCell->flushCPA(ptrCPA);
                        ptrCell->nbcCPADec();
                        nbCPA_--; nbCPAKill++;
                        nbCPAPres_--;
                        CPA_free.push_back(ptrCPA);
                     }
                  }
                  if(ptrCell->getNbcTh()>0) {
                     ptrTh=ptrCell->getTh();
                     if(Th_tab[ptrTh].getInfected()==Tc_tab[ptrTc].getActive()) {
                        nbTh_--;  nbThKill++;
                        nbThInf_--;
                        if(Th_tab[ptrTh].getActive()>0) nbThAct_--;
                        ptrCell->flushTh(ptrTh);
                        ptrCell->nbcThDec();
                        Th_free.push_back(ptrTh);
                     }
                  }
               } //ENDIF_653
	       
               // A passive Tc can get activated or reactivated (memory cells)
               else { // IF_676
	       
                  // (Re)activation by a Th agent
                  if(ptrCell->getNbcTh()>0) {
                     ptrTh=ptrCell->getTh();
		     
                     // First activation by an active Th agent
                     if(Th_tab[ptrTh].getActive()>0) {
                        if(Th_tab[ptrTh].getClonotype()==Tc_tab[ptrTc].getClonotype()) {
                           Tc_tab[ptrTc].setActive(Th_tab[ptrTh].getActive());
                           Tc_tab[ptrTc].setExpansion(1);
                           nbTcAct++;nbTcVerif_++;
                        }
                     }
		     
                     // Reactivation by an infected Th agent
                     else if((Tc_tab[ptrTc].getMemory()>0)&&(Th_tab[ptrTh].getInfected()==Tc_tab[ptrTc].getMemory())) {
                        Tc_tab[ptrTc].setActive(Th_tab[ptrTh].getInfected());
                        Tc_tab[ptrTc].setExpansion(1);
                        nbTcAct++;nbTcVerif_++;
                     }
                  }
		  
                  // Reactivation by an APC agent
                  if((Tc_tab[ptrTc].getActive()==0)&&(Tc_tab[ptrTc].getMemory()>0)&&(ptrCell->getNbcCPA()>0)) {
                     ptrCPA=ptrCell->getCPA();
                     presenting=CPA_tab[ptrCPA].getPresenting();
                     if(presenting==Tc_tab[ptrTc].getMemory()) {
                        Tc_tab[ptrTc].setActive(presenting);
                        Tc_tab[ptrTc].setExpansion(1);
                        nbTcAct++;nbTcVerif_++;
                     }
                  }
		  
                  // Reactivation by a virion
                  if((Tc_tab[ptrTc].getActive()==0)&&(Tc_tab[ptrTc].getMemory()>0)&&(ptrCell->getNbcV()>0)) {
                     ptrV=ptrCell->getV();
                     if(V_tab[ptrV].getStrain()==Tc_tab[ptrTc].getMemory()) {
                        Tc_tab[ptrTc].setActive(V_tab[ptrV].getStrain());
                        Tc_tab[ptrTc].setExpansion(1);
                        nbTcAct++;nbTcVerif_++;
                     }
                  }
               } // ENDIF_676
            } // ENDIF_613
         } // ENDIF_576

         // If a Tc agent is too old, it either dies or become a memory cell
         else { // IF_726

            if(Tc_tab[ptrTc].getMemory()==0) {

               // 95% of Tc agents die
               if(random_int(1,100)>5) {
                  if(Tc_tab[ptrTc].getActive()>0) nbTcAct_--;
                  nbTc_--;  nbTcKill++;
                  ptrCell->flushTc(ptrTc);
                  ptrCell->nbcTcDec();
                  Tc_free.push_back(ptrTc);
               }
	       
               // 5% of them become memory cells
               else {
                  Tc_tab[ptrTc].setMemory(Tc_tab[ptrTc].getActive());
                  Tc_tab[ptrTc].setActive(0);
                  Tc_tab[ptrTc].setMaxAge(10*MAXAGETC);
               }
            }
	    
            // Old memory cells just die
            else {
               if(Tc_tab[ptrTc].getActive()>0) nbTcAct_--;
               nbTc_--;  nbTcKill++;
               ptrCell->flushTc(ptrTc);
               ptrCell->nbcTcDec();
               Tc_free.push_back(ptrTc);
            }
         } // ENDIF_726
      } // ENDIF_570

      // Simulating an action from one CPA agent of the matrix element
      if(ptrCell->getNbcCPA()>0) { // IF_758
         // The pointer is set on an APC agent randomly chosen
         ptrCPA=ptrCell->getCPA();

         // If the agent is still alive, it can move and perform actions
         if(current-CPA_tab[ptrCPA].getDate()<CPA_tab[ptrCPA].getMaxAge()) { // IF_764

            // The next possible destination is computed
            i = getI(cpt); j = getJ(cpt);
            move=CPA_tab[ptrCPA].getMove();
            i = getMoveI( i, move ); j = getMoveJ( j, move );
            newCell=(i*size_+j)%(size_*size_);
            if(newCell<0) newCell+=(size_*size_);
            ptrCellVois=&myMatrix_[newCell];
            temp=CPA_tab[ptrCPA].getPresenting();

            // The agent is moved if there is enough room left
            if(ptrCellVois->getNbcCPA()<NBCPACELL) { // si il reste de la place
               moveCPA(ptrCPA,ptrCell,ptrCellVois);
               ptrCell=&myMatrix_[newCell%(size_*size_)];
            }
            else newCell=cpt;

            // The agent leaves the node if it moved to the exit point
            // The second parameter is the type of agent (3 is for APC agents)
            if(newCell==(size_*size_-1)) { // IF_784
               if(CPA_tab[ptrCPA].getDest()<0) CPA_tab[ptrCPA].setDest(random_int(0,NBNOEUDS-1));
               tempRandom=random_int(0,100);
               if((wd->sendTo[1]>=0)&&((wd->appartient(wd->route,CPA_tab[ptrCPA].getDest()))||(tempRandom>=50)))
                  agentDepartSuccess=wd->addMovingAgent_list2(CPA_tab[ptrCPA].getDest(),3,
                               CPA_tab[ptrCPA].getDate(),
                               CPA_tab[ptrCPA].getMaxAge(),
                               CPA_tab[ptrCPA].getPresenting1(),
                               CPA_tab[ptrCPA].getPresenting2(),
                               CPA_tab[ptrCPA].getPresenting3(),
                               CPA_tab[ptrCPA].getPresenting4(),
                               CPA_tab[ptrCPA].getPresenting5());
               else
                  agentDepartSuccess=wd->addMovingAgent_list1(CPA_tab[ptrCPA].getDest(),3,
                               CPA_tab[ptrCPA].getDate(),
                               CPA_tab[ptrCPA].getMaxAge(),
                               CPA_tab[ptrCPA].getPresenting1(),
                               CPA_tab[ptrCPA].getPresenting2(),
                               CPA_tab[ptrCPA].getPresenting3(),
                               CPA_tab[ptrCPA].getPresenting4(),
                               CPA_tab[ptrCPA].getPresenting5());

               // The exit point is updated
               if(agentDepartSuccess==0) {
                  ptrCellVois->flushCPA(ptrCPA);
                  ptrCellVois->nbcCPADec();
                  nbCPA_--; nbCPASent++;
                  CPA_free.push_back(ptrCPA);
               }
            } // ENDIF_784

            // If it moved to a standard element, it can perform actions
            else { // IF_798

               // If it is not presenting an antigen yet, it scans for virions
               presenting=CPA_tab[ptrCPA].getPresenting();
               if((presenting==0) && (ptrCell->getNbcV()>0)) {
                  ptrV=ptrCell->getV();
                  CPA_tab[ptrCPA].setPresenting(V_tab[ptrV].getStrain());
                  nbCPAPres_++;nbCPAVerif_++;
                  ptrCell->flushV(ptrV);
                  nbV_--;  nbVKill++; nbCPAPres++;
                  ptrCell->nbcVDec();
                  V_free.push_back(ptrV);
               }

               // It also scans for infected CD4
               presenting=CPA_tab[ptrCPA].getPresenting();
               if((presenting==0) && (ptrCell->getNbcTh()>0)) {
                  ptrTh=ptrCell->getTh();
                  CPA_tab[ptrCPA].setPresenting(Th_tab[ptrTh].getInfected());
                  nbCPAPres_++;nbCPAVerif_++;
                  nbCPAPres++;
               }
            } // ENDIF_798
         } // ENDIF_764

         // If the agent is too old, it dies
         else {

            if((CPA_tab[ptrCPA].getPresenting1()>0) ||
               (CPA_tab[ptrCPA].getPresenting2()>0) ||
               (CPA_tab[ptrCPA].getPresenting3()>0) ||
               (CPA_tab[ptrCPA].getPresenting4()>0) ||
               (CPA_tab[ptrCPA].getPresenting5()>0))   nbCPAPres_--;
               nbCPA_--; nbCPAKill++;
               ptrCell->flushCPA(ptrCPA);
               ptrCell->nbcCPADec();
               CPA_free.push_back(ptrCPA);
         }
      } // ENDIF_758
   } // ENDIF_388
   // The iteration is over...
   
   // New cells created (e.g. by thymus for T cells) are added to a node
   // This will be replaced by an explicit implementation of related organs
   // (see Thesis for details)
   
   // Input of new cells is a function of local cell depletion
   int thymusInputCD4=0, thymusInputCD8=0, inputAPC=0;
   if(nbTh_>nbTh_Init) thymusInputCD4=1;
   else {
      thymusInputCD4=2;
      if(nbTh_<(nbTh_Init/2)) thymusInputCD4*=2;
      if(nbTh_<(nbTh_Init/3)) thymusInputCD4*=2;
   }
   if(nbTc_>nbTc_Init) thymusInputCD8=1;
   else {
      thymusInputCD8=2;
      if(nbTc_<(nbTc_Init/2)) thymusInputCD8*=2;
      if(nbTc_<(nbTc_Init/3)) thymusInputCD8*=2;
   }
   if(nbCPA_>nbCPA_Init) inputAPC=1;
   else {
      inputAPC=2;
      if(nbCPA_<(nbCPA_Init/2)) inputAPC*=2;
      if(nbCPA_<(nbCPA_Init/3)) inputAPC*=2;
   }

   // Th cells are added
   for(int k=0; k<thymusInputCD4; ++k) {
      int cell_i = random_int(0,size_-1);
      int cell_j = random_int(0,size_-1);
      int thereIsNoRoom=0;
      do {
         ptrCell = access(cell_i,cell_j);
         ++cell_i;
         if(cell_i==size_) {cell_i=0; ++cell_j; if(cell_j==size_) cell_j=0;}
         ++thereIsNoRoom;
      }
      while(ptrCell->getNbcTh()==NBTHCELL && thereIsNoRoom<size_*size_);
      if(thereIsNoRoom<size_*size_) {
         ptrTh = Th_free.back();
         Th_free.pop_back();
         ptrCell->tabTh[ptrCell->getNbcTh()]=ptrTh;
         Th_tab[ptrTh].setDate(current);
         Th_tab[ptrTh].setDest(-1);
         Th_tab[ptrTh].setMaxAge(MAXAGETH);
         Th_tab[ptrTh].setClonotype(random_int(0,EPITOPE_NUMBER));
         Th_tab[ptrTh].setActive(0);
         Th_tab[ptrTh].setInfected(0);
         Th_tab[ptrTh].setMemory(0);
         Th_tab[ptrTh].setExpansion(0);
         ptrCell->nbcThInc(); nbTh_++;  nbThCrea++;
      }
   }
   
   // Tc cells are added
   for(int k=0; k<thymusInputCD8; ++k) {
      int cell_i = random_int(0,size_-1);
      int cell_j = random_int(0,size_-1);
      int thereIsNoRoom=0;
      do {
         ptrCell = access(cell_i,cell_j);
         ++cell_i;
         if(cell_i==size_) {cell_i=0; ++cell_j; if(cell_j==size_) cell_j=0;}
         ++thereIsNoRoom;
      }
      while(ptrCell->getNbcTc()==NBTCCELL && thereIsNoRoom<size_*size_);
      if(thereIsNoRoom<size_*size_) {
         ptrTc = Tc_free.back();
         Tc_free.pop_back();
         ptrCell->tabTc[ptrCell->getNbcTc()]=ptrTc;
         Tc_tab[ptrTc].setDate(current);
         Tc_tab[ptrTc].setDest(-1);
         Tc_tab[ptrTc].setMaxAge(MAXAGETC);
         Tc_tab[ptrTc].setClonotype(random_int(0,EPITOPE_NUMBER));
         Tc_tab[ptrTc].setActive(0);
         Tc_tab[ptrTc].setExpansion(0);
         Tc_tab[ptrTc].setMemory(0);
         ptrCell->nbcTcInc(); nbTc_++;  nbTcCrea++;
      }
   }
   
   // APC cells are added
   for(int k=0; k<inputAPC; ++k) {
      int cell_i = random_int(0,size_-1);
      int cell_j = random_int(0,size_-1);
      int thereIsNoRoom=0;
      do {
         ptrCell = access(cell_i,cell_j);
         ++cell_i;
         if(cell_i==size_) {cell_i=0; ++cell_j; if(cell_j==size_) cell_j=0;}
         ++thereIsNoRoom;
      }
      while(ptrCell->getNbcTc()==NBCPACELL && thereIsNoRoom<size_*size_);
      if(thereIsNoRoom<size_*size_) {
         ptrCPA = CPA_free.back();
         CPA_free.pop_back();
         ptrCell->tabCPA[ptrCell->getNbcCPA()]=ptrCPA;
         CPA_tab[ptrCPA].setDate(current);
         CPA_tab[ptrCPA].setDest(-1);
         CPA_tab[ptrCPA].setMaxAge(MAXAGECPA);
         CPA_tab[ptrCPA].setPresenting1(0);
         CPA_tab[ptrCPA].setPresenting2(0);
         CPA_tab[ptrCPA].setPresenting3(0);
         CPA_tab[ptrCPA].setPresenting4(0);
         CPA_tab[ptrCPA].setPresenting5(0);
         ptrCell->nbcCPAInc(); nbCPA_++;  nbCPACrea++;
      }
   } // ENDIF_829
   
}






/***************************************************************
*                                                              *
*    Node::VirusRecog()                                        *
*	   returns true if the strain given as a parameter     *
*          is recognized by the clonotype given                 *
*                                                              *
***************************************************************/

bool Node::VirusRecog(int strain, int clonotype) {
   bool val=false;

   // Dichotomic search in the vector "certain" of the strain   
   int a=0; int b=strains[strain].certain.size()-1; int m;
   while((a<b)&&(val==false)) {
      m=(a+b)/2;
      if(strains[strain].certain[m]<clonotype) a=m+1;
      else {
         if (strains[strain].certain[m]>clonotype) b=m-1;
         else {
            if(strains[strain].certain[m]==clonotype) {
               val=true;
               // "clonotype" is in the "certain" vector at position m
            }
         }
      }
   }
   
   if(val!=true) { // "clonotype" is not in the certain vector  
      // Dichotomic search in the vector "random" of the strain   
      int a=0; int b=strains[strain].random.size()-1; int m;
      while((a<b)&&(val==false)) {
         m=(a+b)/2;
         if(strains[strain].random[m]<clonotype) a=m+1;
         else {
            if (strains[strain].random[m]>clonotype) b=m-1;
            else {
               if(strains[strain].random[m]==clonotype) {
                  val=true;
                  // "clonotype" is in the "random" vector at position m
               }
            }
         }
      }

      if(val==true) { // "clonotype" has to be moved from "random" to "certain"

         // Added to "certain"
         strains[strain].certain.push_back(clonotype);
         std::sort(strains[strain].certain.begin(), strains[strain].certain.end());
	 
         // Removed from "random"
         int i=0;
         vector<int>::iterator it;
         for(it=strains[strain].random.begin();it!=strains[strain].random.end();++it) {
            if(strains[strain].random[i]==clonotype) strains[strain].random.erase(it);
            ++i;
         }
      }
   }
   return val;
}


/* ------------------------------ End Of File ------------------------------- */
