#ifndef NODE_H
#define NODE_H

#include "cellule.hpp"
#include "strain.hpp"
//#include "world.hpp"
#include <vector>

#define STRAIN_NUMBER 500

class World;

class Node
{
   public :
   
      Node();
      virtual ~Node();
      bool init(int, int, int, int, int);
      void initStrains();
      void iteration(int, World*,int);


   public :

      Cellule* access(int inX, int inY) 
         {return &(myMatrix_[inX * size_ + inY]);}

      inline int getMoveI(int inI, int inMove){
         static int di[8]={0,1,1,1,0,-1,-1,-1} ;
         int tmpI;
         tmpI=(inI+di[inMove]);
         if(tmpI==-1) tmpI=1;           // non-periodic borders
         if(tmpI==size_) tmpI=size_-2;
         return tmpI;
      };
     
      inline int getMoveJ(int inJ, int inMove) {
         static int dj[8]={-1,-1,0,1,1,1,0,-1} ;
         int tmpJ;
         tmpJ=(inJ+dj[inMove]);
         if(tmpJ==-1) tmpJ=1;           // non-periodic borders
         if(tmpJ==size_) tmpJ=size_-2;
         return tmpJ;
      };
     
      inline int getI(int pos) {return pos/size_;};
      inline int getJ(int pos) {return pos%size_;};
     
      void moveV(int,Cellule*,Cellule*);
      void moveTh(int,Cellule*,Cellule*);     
      void moveTc(int,Cellule*,Cellule*);
      void moveCPA(int,Cellule*,Cellule*);
      bool VirusRecog(int strain, int epitope);

   public :
      int size_;
      Cellule *myMatrix_;
      Virus* V_tab;
      vector<int> V_free;
      Th* Th_tab;
      vector<int> Th_free;
      Tc* Tc_tab;
      vector<int> Tc_free;
      CPA* CPA_tab;
      vector<int> CPA_free;
      
      int procNum;

      Strain strains[STRAIN_NUMBER];
     
      int nbV_, nbTh_, nbThInf_, nbThAct_, nbTc_, nbTcAct_, nbCPA_, nbCPAPres_;
      int nbV_Init, nbTh_Init, nbTc_Init, nbCPA_Init;
      int nbTcVerif_, nbCPAVerif_;
      int nbVCrea, nbVKill, nbVRec, nbVSent;
      int nbThInf, nbThAct, nbThCrea, nbThKill, nbThRec, nbThSent;
      int nbTcAct, nbTcCrea, nbTcKill, nbTcRec, nbTcSent;
      int nbCPAPres, nbCPACrea, nbCPAKill, nbCPARec, nbCPASent;
};

#endif // NODE_H


/* ------------------------------ End Of File ------------------------------- */
