#include "defines.h"

#include <stdio.h>
#include <limits.h>
#include <assert.h>


/* ====================================================================== */
/* change the state of a node                                             */
/* complexity: 0(max{a,b})                                                */
/* ====================================================================== */

/* inline */ void changeState (Problem * pb, uint16 node)
{
  assert(node < pb->nb_nodes);

  if (node < pb->nb_genes) // modif on a gene
  {
    if (pb->used[node] == false)
    {
      // add it
      for (uint16 j = 0; j < pb->nb_cond; j++)
      {
        if (pb->used_c[j] == true)
          pb->cost_c[j] -= pb->C[node][j];
        else
          pb->cost_c[j] += pb->C[node][j];
      }
      pb->nb_active_g++;
    }
    else
    {
      // drop it
      for (uint16 j = 0; j < pb->nb_cond; j++)
      {
        if (pb->used_c[j] == true)
          pb->cost_c[j] += pb->C[node][j];
        else
          pb->cost_c[j] -= pb->C[node][j];
      }
      pb->nb_active_g--;
    }
  }
  else // modif on a condition
  {
    uint16 elt = node - pb->nb_genes;
    if (pb->used[node] == false)
    {
      // add it
      for (uint16 i = 0; i < pb->nb_genes; i++)
      {
        if (pb->used_g[i] == true)
          pb->cost_g[i] -= pb->C[i][elt];
        else
          pb->cost_g[i] += pb->C[i][elt];
      }
      pb->nb_active_c++;
    }
    else
    {
      // drop it
      for (uint16 i = 0; i < pb->nb_genes; i++)
      {
        if (pb->used_g[i] == true)
          pb->cost_g[i] += pb->C[i][elt];
        else
          pb->cost_g[i] -= pb->C[i][elt];
      }
      pb->nb_active_c--;
    }
  }

  // update the informations of the elt
  pb->used[node] = !(pb->used[node]);
  pb->value += pb->cost[node];
  pb->cost[node] = - pb->cost[node];

  checkGraphSolution (pb);
}
