#include "defines.h"

#include <stdio.h>
#include <limits.h>
#include <assert.h>



/* ====================================================================== */
/* basic local search with Swap moves on the conditions only              */
/* only allow improvements, moves imply modification on the genes         */
/* TODO: improve the evaluation of the Swap                               */
/* ====================================================================== */


void LS6 (Problem * pb, Solution * s)
{
  // quick check
  if ((s->nb_active_c == 0) || (s->nb_active_c == pb->nb_cond))
    return;

  // initialization
  insertSolution (pb, s);
  checkGraphSolution (pb);

//printf("              init:   val = %d, %d conditions\n", pb->value, pb->nb_active_c);

  // improvement loop
  for (uint16 it = 0; ; it++)
  {
    // Swap mvt: modif of 1+1 conditions + modif on the genes
    int    best_val  = 1;
    uint16 best_add  = INT_MAX;
    uint16 best_drop = INT_MAX;
    for (uint16 j = 0; j < pb->nb_cond; j++)
    {
      // quick check
      if (pb->used_c[j] == true)
        continue;

      for (uint16 k = 0; k < pb->nb_cond; k++)
      {
        if (pb->used_c[k] == false)
          continue;

        int value = 0;
        for (uint16 i = 0; i < pb->nb_genes; i++)
        {
          int delta = pb->C[i][j] - pb->C[i][k];
          if (pb->used_g[i] == true)
          {
            if (delta < pb->cost_g[i])
              value += delta;  // keep and update
            else
              value += pb->cost_g[i]; // drop with new cost
          }
          else
          {
            if (pb->cost_g[i] + delta < 0)
              value += pb->cost_g[i] + delta; // add
          }
        }

        if (value < best_val)
        {
          best_val = value;
          best_add = j;
          best_drop = k;
//printf("best val = %d, add = %d, drop = %d\n", best_val, best_add, best_drop);
        }
      }
    }

    // check the end of the local search
    if (best_val >= 0)
      break;

    // perform the modification
    int val = pb->value;
    changeState (pb, pb->nb_genes + best_add);
    changeState (pb, pb->nb_genes + best_drop);
    for (uint16 i = 0; i < pb->nb_genes; i++)
    {
      if (pb->cost_g[i] < 0)
        changeState (pb, i);
    }
    assert(val + best_val == pb->value);
// printf("              it %3d: val = %d, %d conditions (expected %d)\n", it, pb->value, pb->nb_active_c, best_val);
  }

  // save the solution
  extractSolution (pb, s);
}
