#include "defines.h"

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <mpi.h>

// test on genetic algorithm

void testDimitriGA (Problem * pb)
{
  double start, stop;
  Solution sol;

  // initialization
  allocSolution (&sol, pb->nb_nodes);

  // GA
  {
    start = startChrono ();
    genetic (pb, pb->nb_cond, &sol, false);
    stop = stopChrono (start);
    int procNum;
    MPI_Comm_rank(MPI_COMM_WORLD,&procNum);
    if(procNum==0) printf("GA: in %.2f s\n", stop);
  }
}




int main (int argc, char * argv[])
{
  Problem pb;
  readProblem (&pb, argv[1]);

  MPI_Init(&argc, &argv);
  int procNum, clusterSize;
  MPI_Comm_rank(MPI_COMM_WORLD,&procNum);
  MPI_Comm_size(MPI_COMM_WORLD,&clusterSize);
  
  if(procNum==0) printf("%d : %d genes, %d conditions\n", procNum, pb.nb_genes, pb.nb_cond);

  // test on genetic algorithm
  testDimitriGA (&pb);

  // memory freeing
  freeProblem (&pb);

  MPI_Finalize();

  return EXIT_SUCCESS;
}
