#include "defines.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/time.h>
#include <sys/resource.h>


/* ====================================================================== */
/* display an error log then stops the program (vararg)                   */
/* ====================================================================== */

void error (char const file[], uint16 line, char const func[], char const msg[], ...)
{
  va_list list;

  fprintf(stderr, "error %s:%s:%d: ", file, func, line);

  va_start(list, msg);
  vfprintf(stderr, msg, list);
  va_end(list);

  fprintf(stderr, "\n");
  fflush(stdout);

  exit(EXIT_FAILURE);
}


/* ====================================================================== */
/* return the current time in seconds                                     */
/* ====================================================================== */

double startChrono ()
{
  struct rusage time;

  getrusage(RUSAGE_SELF, &time);
  return (time.ru_utime.tv_sec + 1E-6 * time.ru_utime.tv_usec);
}


/* ====================================================================== */
/* compute the elapsed time in seconds                                    */
/* ====================================================================== */

double stopChrono (double start)
{
  struct rusage time;

  getrusage(RUSAGE_SELF, &time);
  return (time.ru_utime.tv_sec + 1E-6 * time.ru_utime.tv_usec - start);
}
